@@ -5,9 +5,9 @@ from __future__ import annotations
 Modern Robotics: Mechanics, Planning, and Control.
 Code Library
 ***************************************************************************
-Author: Huan Weng, Bill Hunt, Jarvis Schultz, Mikhail Todes,
-Email: huanweng@u.northwestern.edu
-Date: January 2018
+Author: Huan Weng, Bill Hunt, Jarvis Schultz, Mikhail Todes, Minchang Sung
+Email: tjdalsckd@gmail.com
+Date: January 2023
 ***************************************************************************
 Language: Python
 Also available in: MATLAB, Mathematica
@@ -55,7 +55,17 @@ def Normalize(V):
 '''
 *** CHAPTER 3: RIGID-BODY MOTIONS ***
 '''
-
+def EulerXYZ(a,b,r):
+    c_a = np.cos(a)
+    c_b = np.cos(b)
+    c_r = np.cos(r)
+    s_a = np.sin(a)
+    s_b = np.sin(b)
+    s_r = np.sin(r) 
+    R = np.array([[c_a*c_b , c_a*s_b*s_r - s_a*c_r, c_a*s_b*c_r+s_a*s_r],
+                  [s_a*c_b , s_a*s_b*s_r + c_a*c_r, s_a*s_b*c_r-c_a*s_r],
+                  [-s_b, c_b*s_r , c_b*c_r]])
+    return R;
 def RotInv(R):
     """Inverts a rotation matrix
 
@@ -1575,13 +1585,18 @@ def CubicTimeScalingDot(Tf, t):
     :return: The path parameter sdot(t) corresponding to a third-order
              polynomial motion that begins and ends at zero velocity
 
-    Example Input:
-        Tf = 2
-        t = 0.6
-    Output:
-        0.216
     """
-    return 2*3 * (1.0 * t / Tf)/ Tf  - 2 *3* (1.0 * t / Tf) ** 2/ Tf
+    return 3*2/ Tf * (1.0 * t / Tf) - 2 *3/ Tf* (1.0 * t / Tf) ** 2
+def CubicTimeScalingDdot(Tf, t):
+    """Computes sddot(t) for a cubic time scaling
+
+    :param Tf: Total time of the motion in seconds from rest to rest
+    :param t: The current time t satisfying 0 < t < Tf
+    :return: The path parameter sddot(t) corresponding to a third-order
+             polynomial motion that begins and ends at zero velocity
+
+    """
+    return 3*2/ Tf * (1.0 * 1.0 / Tf) - 2 *3/ Tf*2.0/ Tf* (1.0 * t / Tf)    
 def QuinticTimeScaling(Tf, t):
     """Computes s(t) for a quintic time scaling
 
@@ -1614,8 +1629,17 @@ def QuinticTimeScalingDot(Tf, t):
     Output:
         0.16308
     """
-    return 10*3 * (1.0 * t / Tf) ** 2/ Tf - 15 * 4* (1.0 * t / Tf) ** 3/ Tf \
-           + 6 *5 (1.0 * t / Tf) ** 4/ Tf
+    return 10*3.0/Tf*(t/Tf)**2 -15.0*4.0 /Tf*(t/Tf)**3 + 6.0/Tf*5.0*(t/Tf)**4
+def QuinticTimeScalingDdot(Tf, t):
+    """Computes sddot(t) for a quintic time scaling
+
+    :param Tf: Total time of the motion in seconds from rest to rest
+    :param t: The current time t satisfying 0 < t < Tf
+    :return: The path parameter sddot(t) corresponding to a fifth-order
+             polynomial motion that begins and ends at zero velocity and zero
+             acceleration
+    """
+    return 10*3.0/Tf/Tf*2.0*(t/Tf) -15.0*4.0*3.0 /Tf/Tf*(t/Tf)**2 + 6.0/Tf*5.0*4.0/Tf*(t/Tf)**3
 def JointTrajectory(thetastart, thetaend, Tf, N, method):
     """Computes a straight-line trajectory in joint space
 
@@ -1716,74 +1740,11 @@ def ScrewTrajectory(Xstart, Xend, Tf, N, method):
         = np.dot(Xstart, MatrixExp6(MatrixLog6(np.dot(TransInv(Xstart), \
                                                       Xend)) * s))
     return traj
-def CartesianVelTrajectory(Xstart, Xend, Tf, N, method):
-    """Computes a trajectory as a list of N SE(3) matrices corresponding to
-    the origin of the end-effector frame following a straight line
 
-    :param Xstart: The initial end-effector configuration
-    :param Xend: The final end-effector configuration
-    :param Tf: Total time of the motion in seconds from rest to rest
-    :param N: The number of points N > 1 (Start and stop) in the discrete
-              representation of the trajectory
-    :param method: The time-scaling method, where 3 indicates cubic (third-
-                   order polynomial) time scaling and 5 indicates quintic
-                   (fifth-order polynomial) time scaling
-    :return: The discretized trajectory as a list of N matrices in SE(3)
-             separated in time by Tf/(N-1). The first in the list is Xstart
-             and the Nth is Xend
-    This function is similar to ScrewTrajectory, except the origin of the
-    end-effector frame follows a straight line, decoupled from the rotational
-    motion.
-
-    Example Input:
-        Xstart = np.array([[1, 0, 0, 1],
-                           [0, 1, 0, 0],
-                           [0, 0, 1, 1],
-                           [0, 0, 0, 1]])
-        Xend = np.array([[0, 0, 1, 0.1],
-                         [1, 0, 0,   0],
-                         [0, 1, 0, 4.1],
-                         [0, 0, 0,   1]])
-        Tf = 5
-        N = 4
-        method = 5
-    Output:
-        [np.array([[1, 0, 0, 1]
-                   [0, 1, 0, 0]
-                   [0, 0, 1, 1]
-                   [0, 0, 0, 1]]),
-         np.array([[ 0.937, -0.214,  0.277, 0.811]
-                   [ 0.277,  0.937, -0.214,     0]
-                   [-0.214,  0.277,  0.937, 1.651]
-                   [     0,      0,      0,     1]]),
-         np.array([[ 0.277, -0.214,  0.937, 0.289]
-                   [ 0.937,  0.277, -0.214,     0]
-                   [-0.214,  0.937,  0.277, 3.449]
-                   [     0,      0,      0,     1]]),
-         np.array([[0, 0, 1, 0.1]
-                   [1, 0, 0,   0]
-                   [0, 1, 0, 4.1]
-                   [0, 0, 0,   1]])]
-    """
-    N = int(N)
-    timegap = Tf / (N - 1.0)
-    traj = [[None]] * N
-    Rstart, pstart = TransToRp(Xstart)
-    Rend, pend = TransToRp(Xend)
-    for i in range(N):
-        if method == 3:
-            sdot = CubicTimeScalingDot(Tf, timegap * i)
-        else:
-            sdot = QuinticTimeScalingDot(Tf, timegap * i)
-        traj[i] \
-        = np.r_[np.c_[np.dot(Rstart, \
-        np.dot(MatrixLog3(np.dot(np.array(Rstart).T,Rend)),MatrixExp3(MatrixLog3(np.dot(np.array(Rstart).T,Rend))*s) * sdot)), \
-                   sdot * np.array(pend)  - sdot* np.array(pstart)], \
-                   [[0, 0, 0, 1]]]
-    return traj
-def CartesianTrajectory(Xstart, Xend, Tf, N, method):
+def CartesianTrajectory(Xstart, Xend, Tf, N, method,type=0):
     """Computes a trajectory as a list of N SE(3) matrices corresponding to
     the origin of the end-effector frame following a straight line
+    ** IMPROVE VERSION**
 
     :param Xstart: The initial end-effector configuration
     :param Xend: The final end-effector configuration
@@ -1793,6 +1754,7 @@ def CartesianTrajectory(Xstart, Xend, Tf, N, method):
     :param method: The time-scaling method, where 3 indicates cubic (third-
                    order polynomial) time scaling and 5 indicates quintic
                    (fifth-order polynomial) time scaling
+    :parma type : 0--pose, 1--twist, 2--time derivative of twist
     :return: The discretized trajectory as a list of N matrices in SE(3)
              separated in time by Tf/(N-1). The first in the list is Xstart
              and the Nth is Xend
@@ -1838,13 +1800,32 @@ def CartesianTrajectory(Xstart, Xend, Tf, N, method):
     for i in range(N):
         if method == 3:
             s = CubicTimeScaling(Tf, timegap * i)
+            sdot = CubicTimeScalingDot(Tf, timegap * i)
+            sddot = CubicTimeScalingDdot(Tf, timegap * i)          
         else:
             s = QuinticTimeScaling(Tf, timegap * i)
-        traj[i] \
-        = np.r_[np.c_[np.dot(Rstart, \
-        MatrixExp3(MatrixLog3(np.dot(np.array(Rstart).T,Rend)) * s)), \
-                   s * np.array(pend) + (1 - s) * np.array(pstart)], \
-                   [[0, 0, 0, 1]]]
+            sdot = QuinticTimeScalingDot(Tf, timegap * i)
+            sddot = QuinticTimeScalingDdot(Tf, timegap * i) 
+        Rs = np.dot(Rstart, \
+            MatrixExp3(MatrixLog3(np.dot(np.array(Rstart).T,Rend)) * s)) 
+        dRsds =MatrixLog3(np.dot(np.array(Rstart).T,Rend)) @ Rs;
+        ddRsdsds = MatrixLog3(np.dot(np.array(Rstart).T,Rend))@ MatrixLog3(np.dot(np.array(Rstart).T,Rend)) @ Rs;
+        if type ==0:                     
+            traj[i] \
+            = np.r_[np.c_[Rs, \
+                       s * np.array(pend) + (1 - s) * np.array(pstart)], \
+                       [[0, 0, 0, 1]]]
+        elif type==1: 
+            w = so3ToVec((dRsds @ Rs.T )*sdot);
+            v = (np.array(pend)-np.array(pstart))*sdot;
+            Vs = np.r_[w,v]
+            traj[i] = Vs; 
+        elif type==2:
+                  
+            dw = so3ToVec(dRsds @ Rs.T*sddot+(ddRsdsds @ Rs.T + dRsds @ dRsds.T)*sdot*sdot);
+            dv = (np.array(pend)-np.array(pstart))*sddot;
+            dVs = np.r_[dw,dv]
+            traj[i] =dVs;                   
     return traj
 
 '''
